indexing

   description: "Base class that covers Posix file descriptor routines."

   idea: "Functions here are also available under Windows more or less."

   author: "Berend de Boer"
   date: "$Date: Fri, 13 Oct 2000 19:01:32 +0200 $"
   revision: "$Revision: 1.7 $"

-- deferred 
class 

   BASE_FILE_DESCRIPTOR


inherit

   MEMORY
      redefine
         dispose
      end

   POSIX_BASE

   PORTABLE_PATH

   PAPI_STDIO

   PAPI_FCNTL

   PAPI_UNISTD


feature -- creation

   open (a_path: STRING; flags: INTEGER) is
         -- open given file with access given by `flags'
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_open (a_path, flags)
      end

   open_read (a_path: STRING) is
         -- open given file with read-only access
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_open (a_path, O_RDONLY)
      end

   open_write (a_path: STRING) is
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_open (a_path, O_WRONLY)
      end

   open_read_write (a_path: STRING) is
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_open (a_path, O_RDWR)
      end

   open_truncate (a_path: STRING) is
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_open (a_path, O_TRUNC)
      end

   create_read_write (a_path: STRING) is
         -- always create a file, existing or not
         -- give read/write permissions to user only
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_create (a_path, 
                    O_CREAT + O_TRUNC + O_RDWR, 
                    S_IRUSR + S_IWUSR)
      ensure
         -- only user has access (read/write) permissions
      end
   
   create_with_mode (a_path: STRING; flags, mode: INTEGER) is
         -- create a file according to `flags' and with `mode' access
         -- permissions
      require
         closed: is_closed
      do
         make_path
         clear_fd
         do_create (a_path, flags, mode)
      ensure
         -- not is_readable
      end
   
   
feature -- special creation

   attach_to_fd (a_fd: INTEGER) is
         -- Create file descriptor with value `a_fd'
      require
         closed: is_closed
         valid_fd: a_fd >= 0
         -- `a_fd' is open
      do
         make_path
         fd := a_fd
         fd_open := True
      ensure
         opened: is_open
      end

   make_from_file (file: STDC_FILE) is
         -- Create file descriptor from given stream
         -- The stream is leading, so this file descriptor will
         -- never automatically close when garbage collected, but
         -- it will close when `close' is called.
         -- In that case the stream is no longer valid of course, 
         -- but that's up to you to detect.
      require
         closed: is_closed
         valid_file: file /= Void and then file.is_open
      do
         make_path
         fd := posix_fileno (file.stream)
         if fd = -1 then
            raise_posix_error
         end
         fd_open := True
         close_disabled := True
      ensure
         open: is_open
      end

   make_as_duplicate (another: BASE_FILE_DESCRIPTOR) is
         -- On creation, create a duplicate from another file descriptor
         -- As normal call, closes its own descriptor first (if open) and
         -- duplicates next.
      do
         make_path
         if is_open then
            fd := posix_dup2 (another.value, fd)
         else
            fd := posix_dup (another.value)
         end
         if fd = -1 then
            raise_posix_error
         end
         fd_open := True
      ensure
         open: is_open
      end

   
feature {NONE} -- open

   do_open (a_path: STRING; flags: INTEGER) is
      require
         closed: is_closed
         -- no_create_flag: flags && O_CREAT = 0
      local
         cpath: POINTER         
      do
         portable_path.make_expand (a_path)
         cpath := sh.string_to_pointer (portable_path)
         fd := posix_open (cpath, flags)
         if fd = -1 then
            raise_posix_error
         end
         fd_open := True
      ensure
         is_opened: is_open
      end

   do_create (a_path: STRING; flags, mode: INTEGER) is
      require
         closed: not is_open
      local
         cpath: POINTER         
      do
         portable_path.make_expand (a_path)
         cpath := sh.string_to_pointer (portable_path)
         fd := posix_create (cpath, flags, mode)
         if fd = -1 then
            raise_posix_error
         end
         fd_open := True
      ensure
         is_opened: is_open
      end

feature -- close

   close is
         -- we always describe an existing object, however user probably wants 
         -- to have control about closing a file. And because of garbage 
         -- collection we can't free the file_descriptor itself.
      require
         opened: is_open
      do
         if not close_disabled then
            safe_call (posix_close (fd))
         end
         clear_fd
      ensure
         closed: is_closed
      end

   unattach is
         -- unbind from the current file descriptor
      do
         clear_fd
      ensure
         closed: not is_open
      end


feature -- raw read and write

   last_read: INTEGER
         -- last bytes read by read

   read (buf: POINTER; size: INTEGER) is
         -- read data into `buf' for `size' bytes
      require
         valid_buf: buf /= default_pointer
         valid_size: size >= 0
      do
         errno.clear
         last_read := posix_read (fd, buf, size)
         if last_read /= size then
            if errno.value /= 0 then
               debug
                  print ("!!")
                  print (errno.value)
                  print ("%N")
                  print ("!!")
                  print (last_read)
                  print ("%N")
               end
               raise_posix_error
            end
         end
      end

   write (buf: POINTER; size: INTEGER) is
         -- write given data
      require
         valid_buf: buf /= default_pointer
         valid_size: size >= 0
      local
         bytes_written: INTEGER
      do
         bytes_written := posix_write (fd, buf, size)
         if bytes_written /= size then
            raise_posix_error
         end
      end


feature -- Eiffel like read/write
   
   last_string: STRING
         -- last read string (includes %N), see POSIX_TEXT_FILE.chop
   
   read_string (a_size: INTEGER) is
      do
         -- read bytes, make sure there is a trailing character
         -- after the data
         assert_valid_buffer (a_size+1)
         read (read_buffer.ptr, a_size)
         read_buffer.poke_byte (last_read, 0)

         if last_string = Void then
            if last_read > 256 then
               create last_string.make (a_size)
            else
               create last_string.make (256)
            end
         end
         sh.set_string_from_pointer (last_string, read_buffer.ptr)
      end

   write_string (s: STRING) is
      local
         p: POINTER
      do
         p := sh.string_to_pointer (s)
         write (p, s.count)
      end


feature -- file position
   
   seek (offset: INTEGER) is
         -- set file position to given absolute `offset'
      require
         valid_offset: offset >= 0
      do
         safe_call (posix_lseek (fd, offset, SEEK_SET))
      end

   seek_from_current (offset: INTEGER) is
         -- set file position relative to current position
      do
         safe_call (posix_lseek (fd, offset, SEEK_CUR))
      end
   
   seek_from_end (offset: INTEGER) is
         -- set file position relative to end of file
      require
         valid_offset: offset <= 0
      do
         safe_call (posix_lseek (fd, offset, SEEK_END))
      end


feature -- queries

   isatty, is_attached_to_terminal: BOOLEAN is
         -- return true if handle associated with character device
      do
         Result := posix_isatty (fd)
      end

   is_closed: BOOLEAN is
         -- file descriptor is closed?
      do
         Result := not fd_open
      ensure
         in_balance: Result implies not is_open
      end
   
   is_open: BOOLEAN is
         -- still describes a file descriptor?
      do
         Result := fd_open
      ensure
         in_balance: Result implies not is_closed
      end

   status: POSIX_STATUS is
      do
         if my_status = Void then
            create {POSIX_STATUS_FILDES} my_status.make (Current)
         end
         Result := my_status
      end

   value: INTEGER is
         -- return the value of the file descriptor
      require
         valid_file_descriptor: is_open
      do
         Result := fd
      end


feature -- accessible state   
   
   path: STRING is
      do
         Result := portable_path
      end
   

feature {NONE} -- internal

   fd: INTEGER
         -- the actual file descriptor

   fd_open: BOOLEAN
         -- True if fd describes an open file descriptor

   my_status: POSIX_STATUS
         -- cached status object
   

   read_buffer: STDC_DYNAMIC_MEMORY
         -- buffer used by some read routines

   assert_valid_buffer (a_size: INTEGER) is
         -- make sure `read_buffer' can contain at least `a_size'
         -- elements
      require
         a_size >= 0
      do
         if read_buffer = Void then
            create read_buffer.allocate (a_size)
         else
            if read_buffer.size < a_size then
               read_buffer.allocate (a_size)
            end
         end
      ensure
         buffer_large_enough: read_buffer.size >= a_size
      end
   
   clear_fd is
      do
         fd := -1
         fd_open := False
      end

   dispose is
      do
         if is_open then
            close
         end
      end


feature {STDC_FILE} -- prevent subtile garbage collection problems
   
   close_disabled: BOOLEAN
         -- do no longer close the associated file descriptor

   disable_close is
         -- when a stream is opened on a file descriptor the file
         -- descriptor itself should not close itself, the stream
         -- will close it
      do
         close_disabled := True
      end
   

invariant   
   
   valid_internal_file_descriptor: fd >= -1

   
end -- class BASE_FILE_DESCRIPTOR
